import { Vector2 } from 'three'
import { XRUpdateParams } from '../core/XRClass'
import XRController from '../core/XRController'
import ControllerType from '../enums/ControllerType'
import KeyCode from '../enums/KeyCode'
import { MousePos } from '../enums/MousePos'

export default class KeyboardController extends XRController {
	static controllerType = ControllerType.KEYBOARD
	shiftDown: boolean
	mouseDown: boolean
	keysDown: KeyCode[]
	mouseDownPos: MousePos
	mouseUpPos: MousePos
	mousePos: MousePos
	mouseDiff: Vector2

	defaults() {
		return {
			...super.defaults(),
			controllerType: ControllerType.KEYBOARD,
			shiftDown: false,
			mouseDown: false,
			keysDown: [],
			mouseDownPos: new MousePos(),
			mouseUpPos: new MousePos(),
			mousePos: new MousePos(),
			mouseDiff: new Vector2(),
		}
	}

	init() {
		super.init()
		this._onKeyDown = this.onKeyDown.bind(this)
		this._onKeyUp = this.onKeyUp.bind(this)
		this._handleMouseDown = this.handleMouseDown.bind(this)
		this._onMouseMove = this.onMouseMove.bind(this)
		this._onMouseUp = this.onMouseUp.bind(this)
		window.addEventListener('mousemove', this._onMouseMove)
		window.addEventListener('mousedown', this._handleMouseDown)
		window.addEventListener('mouseup', this._onMouseUp)
		window.addEventListener('keydown', this._onKeyDown)
		window.addEventListener('keyup', this._onKeyUp)
	}

	update(p?: XRUpdateParams) {
		this.mouseDiff.set(0, 0)
		super.update(p)
	}

	isKeyDown(keyCode: KeyCode) {
		return this.keysDown.indexOf(keyCode) >= 0
	}

	_handleMouseDown: any
	handleMouseDown(e: MouseEvent) {
		if (e.target !== this.project.viewer.canvas) return
		this.onMouseDown(e)
	}
	onMouseDown(e: MouseEvent) {
		this.mouseDownPos = new MousePos(e)
		this.mouseDown = true
		this.mouseDiff.set(0, 0)
		this.mousePos = new MousePos(e)
	}

	_onMouseUp: any
	onMouseUp(e: MouseEvent) {
		this.mouseUpPos = new MousePos(e)
		this.mouseDown = false
	}

	_onMouseMove: any
	onMouseMove(e: MouseEvent) {
		const newPos = new MousePos(e)
		if (this.mousePos && this.mouseDown)
			this.mouseDiff.add(this.mousePos.client.clone().sub(newPos.client))
		this.mousePos = newPos
	}

	_onKeyDown: any
	onKeyDown(e: KeyboardEvent) {
		this.shiftDown = e.shiftKey
		if (this.isKeyDown(e.keyCode)) return
		if (Object.values(KeyCode).indexOf(e.keyCode) >= 0) this.keysDown.push(e.keyCode)
	}

	_onKeyUp: any
	onKeyUp(e: KeyboardEvent) {
		this.shiftDown = e.shiftKey
		if (!this.isKeyDown(e.keyCode)) return
		if (Object.values(KeyCode).indexOf(e.keyCode) >= 0)
			this.keysDown = this.keysDown.filter(v => v != e.keyCode)
	}

	destroy() {
		window.removeEventListener('mousemove', this._onMouseMove)
		window.removeEventListener('mousedown', this._handleMouseDown)
		window.removeEventListener('mouseup', this._onMouseUp)
		window.removeEventListener('keydown', this._onKeyDown)
		window.removeEventListener('keyup', this._onKeyUp)
		super.destroy()
	}
}
