import { Material, Matrix4, Mesh, MeshBasicMaterial, Quaternion, Vector3 } from 'three'
import MotionType from '../enums/MotionType'
import quaternionA2T from '../utils/quaternionA2T'
import quaternion3T2A from '../utils/quaternionT2A'
import vector3A2T from '../utils/vector3A2T'
import vector3T2A from '../utils/vector3T2A'
import XRClass, { XRUpdateParams } from './XRClass'

export default class XRMass extends XRClass(Mesh) {
	body: Ammo.btRigidBody
	needsUpdate: boolean
	mass: number
	materialStack: Material[]

	defaults() {
		return {
			...super.defaults(),
			hasBody: false,
			mass: 0,
			breakable: false,
			fragmentDepth: 0,
			fractureImpulse: 1,
		}
	}

	init() {
		this.initGeometry()
		this.materialStack = []
		if (!this.material) this.material = new MeshBasicMaterial({ color: 0xffffff })
		this.materialStack.push(this.material)
	}

	initGeometry() {
		// Override with a function that sets geometry
		//
		// e.g. this.geometry = new ThingyGeometry()
	}

	update(p?: XRUpdateParams) {
		super.update(p)
		if (this.body) {
			if (this.body.isKinematicObject() && this.needsUpdate) {
				this._applyThreeToAmmo()
				this.needsUpdate = false
			} else if (!this.body.isStaticObject()) {
				this._applyAmmoToThree()
			}
		}
	}

	setAngularVelocity(x: number | Vector3 = 0, y: number = 0, z: number = 0) {
		let v = x instanceof Vector3 ? vector3T2A(x) : new Ammo.btVector3(x, y, z)
		this.body.setAngularVelocity(v)
		Ammo.destroy(v)
	}

	setLinearVelocity(x: number | Vector3 = 0, y: number = 0, z: number = 0) {
		let v = x instanceof Vector3 ? vector3T2A(x) : new Ammo.btVector3(x, y, z)
		this.body.setLinearVelocity(v)
		Ammo.destroy(v)
	}

	setMaterial(m: Material) {
		this.materialStack.push(m)
		if (Array.isArray(this.material)) {
			if (m.userData?.importance >= this.material[0]?.userData?.importance || 0)
				this.material[0] = m
		} else {
			if (m.userData?.importance >= (this.material?.userData?.importance || 0))
				this.material = m
		}
	}

	removeMaterial(m: Material | Material[], importance: number = 0) {
		this.materialStack = this.materialStack.filter(mm => mm !== m)
		if (this.material === m) {
			if (this.materialStack.length)
				this.material = this.materialStack[this.materialStack.length - 1]
		}
	}

	turnOnTempKinematics() {
		const b = this.body
		this.userData.prevCollisionFlags = b?.getCollisionFlags()
		if (b && b.getCollisionFlags() !== MotionType.Kinematic)
			b.setCollisionFlags(MotionType.Kinematic)
	}

	turnOffTempKinematics() {
		if (this.userData.prevCollisionFlags === MotionType.Kinematic) return
		const b = this.body
		if (!b) return
		this._applyThreeToAmmo()
		b.setCollisionFlags(this.userData.prevCollisionFlags)
		if (this.userData.jiggle) {
			this.setLinearVelocity(0, 1, 0)
		} else {
			this.stopVelocities()
		}
		this.userData.jiggle = false
	}

	stopVelocities() {
		const b = this.body
		const v = new Ammo.btVector3(0, 0, 0)
		b?.setLinearVelocity(v)
		b?.setAngularVelocity(v)
		Ammo.destroy(v)
	}

	private _applyThreeToAmmo() {
		const b = this.body
		const ammoState = b.getMotionState()
		const ammoT = new Ammo.btTransform()
		ammoState.getWorldTransform(ammoT)

		// get position and rotation
		const thisQ = this.getWorldQuaternion(new Quaternion())
		const thisP = this.getWorldPosition(new Vector3())
		const thisS = this.getWorldScale(new Vector3())

		// set scaling
		b.getCollisionShape().setLocalScaling(vector3T2A(thisS))

		// set position and rotation
		ammoT.setOrigin(vector3T2A(thisP))
		ammoT.setRotation(quaternion3T2A(thisQ))

		// set transform
		ammoState.setWorldTransform(ammoT)
		Ammo.destroy(ammoT)
	}

	private _applyAmmoToThree() {
		const ammoState = this.body.getMotionState()
		const ammoT = new Ammo.btTransform()
		ammoState.getWorldTransform(ammoT)

		// get position and rotation
		const physP = ammoT.getOrigin()
		const physQ = ammoT.getRotation()

		// set position and rotation
		const scale = this.body.getCollisionShape().getLocalScaling()
		const thisS = vector3A2T(scale)
		const thisP = vector3A2T(physP)
		const thisQ = quaternionA2T(physQ)
		const thisM = new Matrix4().compose(thisP, thisQ, thisS)
		new Matrix4()
			.copy(this.parent.matrixWorld)
			.invert()
			.multiply(thisM)
			.decompose(this.position, this.quaternion, this.scale)
		Ammo.destroy(ammoT)
	}
}
