import XYWH from '@thats/utils/types/XYWH'
import React, { useContext, useEffect, useMemo, useRef, useState } from 'react'
import XRParam from '../core/XRParam'
import XRProject from '../core/XRProject'
import GrabActionType from '../enums/GrabActionType'
import PlayState from '../enums/PlayState'
import scenes from '../scenes/index'

export interface XRProjectViewerStateProps {
	children: any
	sceneIdx: number
}

export function XRProjectViewerState({ children, sceneIdx }: XRProjectViewerStateProps) {
	const [changed, setChanged] = useState(new Date())
	const project = useRef<XRProject>(new XRProject())

	useEffect(() => () => project.current.destroy(), [])

	useEffect(() => {
		project.current.setScene(scenes[sceneIdx])
		setChanged(new Date())
	}, [sceneIdx])

	const initViewer = useMemo(
		() => (el: HTMLElement, dims: XYWH) => {
			project.current.initViewer(el, dims)
			setChanged(new Date())
		},
		[]
	)

	const setGrabAction = useMemo(
		() => (v: GrabActionType) => {
			project.current.setGrabActionState(v)
			setChanged(new Date())
		},
		[]
	)

	const setGodMode = useMemo(
		() => (v: boolean) => {
			project.current.setGodModeState(v)
			setChanged(new Date())
		},
		[]
	)

	const setParam = useMemo(
		() => (param: XRParam<any>, value: any) => {
			project.current.setParam(param, value)
			setChanged(new Date())
		},
		[]
	)

	const stopXR = useMemo(
		() => () => {
			project.current.endXR()
			setChanged(new Date())
		},
		[]
	)
	const startXR = useMemo(
		() => () => {
			project.current.startXR()
			setChanged(new Date())
		},
		[]
	)

	const play = useMemo(
		() => () => {
			if (project.current.state.play === PlayState.PAUSED) project.current.play()
			else project.current.start()
			setChanged(new Date())
		},
		[]
	)

	const pause = useMemo(
		() => () => {
			project.current.pause()
			setChanged(new Date())
		},
		[]
	)

	const stop = useMemo(
		() => () => {
			project.current.stop()
			setChanged(new Date())
		},
		[]
	)

	const ctx = useMemo(
		() => ({
			project: project.current,
			setParam,
			setGodMode,
			setGrabAction,
			initViewer,
			play,
			pause,
			stop,
			startXR,
			stopXR,
		}),
		// eslint-disable-next-line
		[changed]
	)

	return (
		<XRProjectViewerStateContext.Provider value={ctx}>
			{children}
		</XRProjectViewerStateContext.Provider>
	)
}
export default XRProjectViewerState

const XRProjectViewerStateContext = React.createContext<{
	project: XRProject
	setParam: <T>(param: XRParam<T>, value: T) => void
	setGodMode: (v: boolean) => void
	setGrabAction: (v: GrabActionType) => void
	initViewer: (el: HTMLElement, dims: XYWH) => void
	play: () => void
	pause: () => void
	stop: () => void
	startXR: () => void
	stopXR: () => void
}>(null)

export const useXRProjectViewer = () => useContext(XRProjectViewerStateContext)
